import faust
from faust import App
from faust.topics import Topic, Any, ChannelT
import logging


LOGGER = logging.getLogger(__name__)


class ShortRepartitionTopic(Topic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def derive(self, **kwargs: Any) -> ChannelT:
        current_suffix = kwargs.get("suffix")
        if 'repartition' in current_suffix:
            kwargs['suffix'] = '-'.join(['-GroupBy', current_suffix.lstrip('-').split('-')[-2]])
            kwargs['prefix'] = ''
        return self.derive_topic(**kwargs)


class InstrumentedApp(App):
    """
    Faust app with metric of number of agent exceptions
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conf.Topic = ShortRepartitionTopic

    async def _on_agent_error(self, agent: faust.agents.AgentT, exc: BaseException) -> None:
        self.wrapper.agent_exception(exc)
        await super(InstrumentedApp, self)._on_agent_error(agent, exc)



import logging
import os
import sys
from abc import ABC, abstractmethod

from faust import Record

from nubium_utils import parse_headers
from nubium_utils.metrics import MetricsManager as MetsMan
from .avro_utils import key_serializer
from .instrumented_app import InstrumentedApp
from .faust_runtime_vars import env_vars

LOGGER = logging.getLogger(__name__)


class FaustMessage(Record, serializer='json'):
    key: str
    value: dict
    headers: dict


class FaustAppWrapper(ABC):
    """
    A wrapper around the Faust app so that it may be more easily unit tested.
    """

    def __init__(self, faust_config, avro_client, metrics_manager: MetsMan):
        self.faust_config = faust_config
        self.avro_client = avro_client

        self.app = InstrumentedApp(**self.faust_config)
        self.app.wrapper = self
        self.metrics_manager = metrics_manager
        self.key_serializer = key_serializer(client=self.avro_client)
        self.input_topics = {}
        self.faust_input_topics = {}

        self._init_serializers()
        self._init_records()
        self._init_topics()
        self._init_tables()
        self._init_agents()
        self._init_metrics_pushing()

    def get_schema_path(self, file, schema_path_from_src_root='./schemas'):
        """
        Ensures that the Faust app's schema files are always correctly referenced/loaded regardless of your init path.
        :param file: name of the schema file
        :param schema_path_from_src_root: based on the root of the app, where the schemas folder is located.
        :return: relative path to load the schema at runtime
        """
        runtime_path = os.getcwd()
        app_file_path = os.path.abspath(sys.modules[self.__module__].__file__)
        common_path = os.path.commonpath([runtime_path, app_file_path])
        rel_path = os.path.relpath(common_path, runtime_path)
        return os.path.join(runtime_path, rel_path, schema_path_from_src_root,
                            file)

    def add_input_topic(self, topic_name, value_serializer):
        """
        Aids the 'self._init_message_parser' by adding faust topics to two different topic dicts:
        One dict is for the original input topics (self.input_topics) which is primarily for debugging/unit testing;
        the other (self.faust_input_topics) is for the FaustMessage version of the topic, where '_faust' appended to the
        original topic name, and the dict key is the original topic name so it's easily referenced.
        Basically, this allows you to call the FaustMessage topic version via self.faust_input_topics[orig_topic_name]
        :param topic_name: original topic name
        :type topic_name: str
        :param value_serializer: a FaustSerializer instance for Avro-encoded message values
        :type value_serializer: FaustSerializer
        """
        faust_name = f'{topic_name}_Internal_{self.app.conf.id}'

        input_topic = self.app.topic(
            topic_name,
            key_serializer=self.key_serializer,
            value_serializer=value_serializer)
        self.input_topics[topic_name] = input_topic

        faust_input_topic = self.app.topic(faust_name,
                                           internal=True,
                                           value_type=FaustMessage)
        self._add_input_agent(topic_name, input_topic, faust_input_topic)
        return faust_input_topic

    def _add_input_agent(self, topic_name, input_topic, faust_input_topic):
        @self.app.agent(input_topic, name=f'{topic_name}_agent_internal')
        async def parse_messages(input_messages):
            async for event in input_messages.events():
                headers = parse_headers(event.headers)
                LOGGER.info(f'Message consumed, GUID: {headers["guid"]}')
                await faust_input_topic.send(
                    value=FaustMessage(
                        key=event.key,
                        value=event.value,
                        headers=headers
                    ),
                    key=event.key
                )
            yield None

    def format_headers(self, headers_dict):
        """
        Formats headers for an outgoing message
        :param headers_dict: the headers as a dict
        :type headers_dict: dict
        :return: the dict as a list of (key, b'value') tuples.
        :rtype: list
        """
        return [(key, value.encode()) for key, value in headers_dict.items()]

    @abstractmethod
    def _init_serializers(self):
        pass

    @abstractmethod
    def _init_records(self):
        pass

    @abstractmethod
    def _init_topics(self):
        pass

    @abstractmethod
    def _init_tables(self):
        pass

    @abstractmethod
    def _init_agents(self):
        pass

    def agent_exception(self, exc):
        """
        Increments the message errors metric by one
        :param exc:
        :return:
        """
        self.metrics_manager.inc_message_errors(exc)

    def _init_metrics_pushing(self):
        """
        Defines method for updating metrics from Faust sensor and pushing data
        :return: None
        """
        if env_vars()['DO_METRICS_PUSHING'].lower() == 'true':
            LOGGER.info('Metric pushing to gateway ENABLED')

            @self.app.timer(int(env_vars()['METRICS_PUSH_RATE']))
            async def push_metrics():
                """
                Updates gauges from Faust and pushes data to prometheus pushgateways
                :return: None
                """
                self.set_gauges()
                self.metrics_manager.push_metrics()
        else:
            LOGGER.info('Metric pushing to gateway DISABLED')

    def set_gauges(self):
        for topic, count in self.app.monitor.messages_received_by_topic.items():
            self.metrics_manager.messages_consumed.labels(
                job=self.metrics_manager.job,
                app=self.metrics_manager.app,
                topic=topic
            ).set(count)
        for topic, count in self.app.monitor.messages_sent_by_topic.items():
            self.metrics_manager.messages_produced.labels(
                job=self.metrics_manager.job,
                app=self.metrics_manager.app,
                topic=topic
            ).set(count)



"""
Container for standardized application monitoring gauges
"""
from prometheus_client import Gauge, CollectorRegistry
from os import environ
from .metrics_pusher import MetricsPusher
from nubium_utils.general_runtime_vars import env_vars


class MetricsManager:
    """
    Coordinates Prometheus monitoring for a Kafka client application

    Creates default gauges, and simplifies updates to them
    """

    def __init__(self, job: str = None, app: str = None, metrics_pusher: MetricsPusher = None, registry: CollectorRegistry = None):
        """
        Initializes monitor

        :param job: (str) Unique name of individual application instance
        :param app: (str) Common app name for grouping metrics
        :param registry: (CollectorRegistry) registry for prometheus metrics
        """
        if not job:
            job = environ['HOSTNAME']
        if not app:
            app = environ['APP_NAME']
        if not metrics_pusher:
            metrics_pusher = MetricsPusher()
        if not registry:
            registry = CollectorRegistry()

        self.job = job
        self.app = app
        self.registry = registry

        self.messages_consumed = Gauge('messages_consumed', 'Messages processed since application start', labelnames=['app', 'job', 'topic'], registry=self.registry)
        self.messages_produced = Gauge('messages_produced', 'Messages successfully produced since application start', labelnames=['app', 'job', 'topic'], registry=self.registry)
        self.message_errors = Gauge('message_errors', 'Exceptions caught when processing messages', labelnames=['app', 'job', 'exception'], registry=self.registry)
        self.external_requests = Gauge('external_requests', 'Network calls to external services', labelnames=['app', 'job', 'request_to', 'request_endpoint', 'request_type', 'is_bulk', 'status_code'], registry=self.registry)
        self.metrics_pusher = metrics_pusher

    def inc_messages_consumed(self, number_of_messages, topic):
        """
        Increases the messages_consumed gauge with default labels
        """
        self.messages_consumed.labels(self.app, self.job, topic).inc(number_of_messages)

    def inc_messages_produced(self, number_of_messages, topic):
        """
        Increases the messages_consumed gauge with default labels
        """
        self.messages_produced.labels(self.app, self.job, topic).inc(number_of_messages)

    def inc_message_errors(self, exception):
        """
        Increases the error gauge with default label and label of the exception
        """
        self.message_errors.labels(self.app, self.job, exception.__class__.__name__).inc(1)

    def inc_external_requests(self, request_to=None, request_endpoint=None, request_type=None, is_bulk=0, status_code=200):
        """
        Increases the external requests gauge.
        """
        self.external_requests.labels(self.app, self.job, request_to, request_endpoint, request_type, is_bulk, status_code).inc(1)

    def push_metrics(self):
        self.metrics_pusher.set_metrics_pod_ips()
        self.metrics_pusher.push_metrics(self.registry)



import logging
from .message_utils import produce_message_callback, consume_message_callback
from nubium_utils.metrics import start_pushing_metrics
from .confluent_runtime_vars import env_vars

LOGGER = logging.getLogger(__name__)


def init_ssl_configs():
    """
    Provides producer or consumer SSL configs.
    :return: config params
    :rtype: dict
    """
    if env_vars()['USE_SSL'].lower() == 'true':
        LOGGER.info('Initializing producers and consumers with TLS configurations ENABLED.')
        return {
            "security.protocol": "ssl",
            "ssl.ca.location": env_vars()["SSL_CA_LOCATION"],
            "ssl.certificate.location": env_vars()["SSL_CERTIFICATE_LOCATION"],
            "ssl.key.location": env_vars()["SSL_KEY_LOCATION"]}
    else:
        LOGGER.info('Initializing producers and consumers with TLS configurations DISABLED.')
    return {}


def init_schema_registry_configs():
    """
    Provides the avro schema config
    :return: dict, None
    """
    url = env_vars()['SCHEMA_REGISTRY_URL']
    if url:
        LOGGER.info(f'Using the schema server at {url}')
        return {'schema.registry.url': url}
    else:
        LOGGER.info(f'No schema registry URL provided; schema registry usage DISABLED')
        return None


def init_producer_configs(ssl_configs=None, schema_registry_configs=None):
    """
    Provides producer config.
    :return: config params
    :rtype: dict
    """
    if not ssl_configs:
        ssl_configs = {}
    if not schema_registry_configs:
        schema_registry_configs = {}
    return {
        **ssl_configs,
        **schema_registry_configs,
        "bootstrap.servers": env_vars()['KAFKA_CLUSTER'],
        "on_delivery": produce_message_callback,
        "acks": "all"}


def init_consumer_configs(ssl_configs=None, schema_registry_configs=None):
    """
    Provides consumer config.
    :return: config params
    :rtype: dict
    """
    if not ssl_configs:
        ssl_configs = {}
    if not schema_registry_configs:
        schema_registry_configs = {}
    return {
        **ssl_configs,
        **schema_registry_configs,
        "bootstrap.servers": env_vars()['KAFKA_CLUSTER'],
        "group.id": env_vars()['APP_NAME'],
        "on_commit": consume_message_callback,
        "enable.auto.commit": "false"}


def init_metrics_pushing(metrics_manager):
    if env_vars()['DO_METRICS_PUSHING'].lower() == 'true':
        LOGGER.info('Metric pushing to gateway ENABLED')
        start_pushing_metrics(metrics_manager, int(env_vars()['METRICS_PUSH_RATE']))
    else:
        LOGGER.info('Metric pushing to gateway DISABLED')
